CREATE OR REPLACE FUNCTION set_invoice_amounts() RETURNS TRIGGER AS $$
    DECLARE
        invoice_ids integer[];
        error_ids integer[];
    BEGIN
        IF TG_OP = 'INSERT' THEN
            invoice_ids := ARRAY[NEW.invoice_id];
        ELSIF TG_OP = 'DELETE' THEN
            invoice_ids := ARRAY[OLD.invoice_id];
        ELSIF TG_OP = 'UPDATE' THEN
            invoice_ids := ARRAY[NEW.invoice_id, OLD.invoice_id];
        END IF;

        IF TG_TABLE_NAME != 'invoicing_invoicepayment' THEN
            EXECUTE 'UPDATE ' || substring(TG_TABLE_NAME for length(TG_TABLE_NAME) - 4) || ' i
            SET total_amount = COALESCE(
                (
                    SELECT SUM(l.total_amount)
                    FROM ' || TG_TABLE_NAME || ' l
                    WHERE l.invoice_id = i.id
                    AND l.status = ''success''
                ), 0
            )
            WHERE id = ANY($1);' USING invoice_ids;
        END IF;

        IF TG_TABLE_NAME != 'invoicing_draftinvoiceline' THEN
            EXECUTE 'UPDATE invoicing_invoice i
            SET paid_amount = COALESCE(
                (
                    SELECT SUM(p.amount)
                    FROM invoicing_invoicepayment p
                    WHERE p.invoice_id = i.id
                ), 0
            )
            WHERE id = ANY($1);' USING invoice_ids;

            EXECUTE 'SELECT ARRAY_AGG(id) FROM invoicing_invoice
            WHERE id = ANY($1) AND paid_amount > total_amount AND NOT (paid_amount = 0 and total_amount < 0)'
            INTO error_ids
            USING invoice_ids;
            IF ARRAY_LENGTH(error_ids, 1) > 0 THEN
                RAISE EXCEPTION 'paid_amount is greater than total_amount for invoices %', error_ids;
            END IF;

            EXECUTE 'UPDATE invoicing_invoice
            SET remaining_amount = total_amount - paid_amount
            WHERE id = ANY($1);' USING invoice_ids;
        END IF;

        IF TG_OP = 'DELETE' THEN
            RETURN OLD;
        ELSE
            RETURN NEW;
        END IF;
    END;
$$ LANGUAGE plpgsql;


DROP TRIGGER IF EXISTS set_draftinvoice_line_trg ON invoicing_draftinvoiceline;
CREATE TRIGGER set_draftinvoice_line_trg
    AFTER INSERT OR UPDATE OF total_amount, invoice_id, status OR DELETE ON invoicing_draftinvoiceline
    FOR EACH ROW
    EXECUTE PROCEDURE set_invoice_amounts();


DROP TRIGGER IF EXISTS set_invoice_line_trg ON invoicing_invoiceline;
CREATE TRIGGER set_invoice_line_trg
    AFTER INSERT OR UPDATE OF total_amount, invoice_id, status OR DELETE ON invoicing_invoiceline
    FOR EACH ROW
    EXECUTE PROCEDURE set_invoice_amounts();


DROP TRIGGER IF EXISTS set_invoice_invoicepayment_trg ON invoicing_invoicepayment;
CREATE TRIGGER set_invoice_invoicepayment_trg
    AFTER INSERT OR UPDATE OF amount OR DELETE ON invoicing_invoicepayment
    FOR EACH ROW
    EXECUTE PROCEDURE set_invoice_amounts();

-- old triggers and function
DROP TRIGGER IF EXISTS set_draftinvoice_total_amount_trg ON invoicing_draftinvoiceline;
DROP TRIGGER IF EXISTS set_invoice_total_amount_trg ON invoicing_invoiceline;
DROP FUNCTION IF EXISTS set_invoice_total_amount;
